"""
Robust principle component analysis
-----------------------------------

.. autoclass:: nussl.separation.factorization.RPCA
    :autosummary:

Independent component analysis
------------------------------

.. autoclass:: nussl.separation.factorization.ICA
    :autosummary:

"""

from .rpca import RPCA
from .ica import ICA
