"""
Cluster by inter-phase and inter-level difference
-------------------------------------------------

.. autoclass:: nussl.separation.spatial.SpatialClustering
    :autosummary:

PROJET: Separate via spatial projections
-------------------------------------------------

.. autoclass:: nussl.separation.spatial.Projet
    :autosummary:

DUET
----

.. autoclass:: nussl.separation.spatial.Duet
    :autosummary:

"""

from .spatial_clustering import SpatialClustering
from .projet import Projet
from .duet import Duet
