"""
Deep clustering
---------------

.. autoclass:: nussl.separation.deep.DeepClustering
    :autosummary:

Deep mask estimation
--------------------

.. autoclass:: nussl.separation.deep.DeepMaskEstimation
    :autosummary:

Deep audio estimation
---------------------

.. autoclass:: nussl.separation.deep.DeepAudioEstimation
    :autosummary:

"""

from .deep_clustering import DeepClustering
from .deep_mask_estimation import DeepMaskEstimation
from .deep_audio_estimation import DeepAudioEstimation
